module edetect_gen

  use param  
  use dal

  implicit none

  private

  public gammln, gammp, gammq

contains

!----------------------------------------------------------------
      real function gammq (a,x)
!----------------------------------------------------------------

!  returns the negative logarithm of the 
!  complemente of the incomplete gamma function p(a,x) 

!  author : s.doebereiner date: 11-sep-1996
!  returns the incomplete gamma function p(a,x) 
!  in the maximum-likelihood representation -ln(p) 
!  a and x must be >0 !
!  modified routines gcf and gser from press et al., numerical recipes, 
!   6.2, p.162
!  a                   r4  i   argument
!  x                   r4  i   argument 
!  gammq               r4    o -ln(q(a,x))

!  gammln              sr      returns ln of the gamma function  
!  wrflag              sr      writes to flag output stream
!  wrerr               sr      write error message

      character*8   rtname
      real          a,x,eps,gln,gold
      real          a0,a1,b0,b1,fac,an,ana,anf,g
      real          ap,sum,del,gamser,gammcf
      integer       itmax,n,ierr

      save

      data          rtname /'gammq'/
      data          itmax /1000/
      data          eps  /3.e-7/                

      ierr = 0
      gammq = 0.

      x = max(x,eps) 


      if (x.lt.a+1) then
        gln=gammln(a)
        ap=a
        sum=1./a
        del=sum
        do 111 n=1,itmax
          ap=ap+1.
          del=del*x/ap
          sum=sum+del
          if (abs(del).lt.abs(sum)*eps) go to 101
  111   continue
        ierr = 1 
!       call wrerr(rtname,ierr,'a too large, itmax too small ')

  101   gamser = sum * exp( -x + a*log(x) - gln)
        gammq = -alog(1.0-gamser)
        
      else

        gln=gammln(a)
        gold=0.
        a0=1.
        a1=x
        b0=0.
        b1=1.
        fac=1.

        do 211 n=1,itmax
          an=float(n)
          ana=an-a
          a0=(a1+a0*ana)*fac
          b0=(b1+b0*ana)*fac
          anf=an*fac
          a1=x*a0+anf*a1
          b1=x*b0+anf*b1
          if (a1.ne.0) then
            fac=1./a1
            g=b1*fac
            if (abs((g-gold)/g).lt.eps) go to 201
            gold=g
          endif
  211   continue
        ierr = 1 
!       call wrerr(rtname,ierr,'a too large, itmax too small ')

  201   gammq = x - a*alog(x) + gln - alog(g)

      endif 

      end function gammq

!----------------------------------------------------------------
      real function gammp (a,x)
!----------------------------------------------------------------

!  returns the negative logarithm of the 
!  incomplete gamma function p(a,x) 

!  author : s.doebereiner date: 11-sep-1996
!  returns the incomplete gamma function p(a,x) 
!  in the maximum-likelihood representation -ln(p) 
!  a and x must be >0 !
!  modified routines gcf and gser from press et al., numerical recipes, 
!   6.2, p.162

!  a                   r4  i   argument
!  x                   r4  i   argument 
!  gammp               r4    o -ln(p(a,x))

!  gammln              sr      returns ln of the gamma function  
!  wrflag              sr      writes to flag output stream
!  wrerr               sr      write error message

      character*8   rtname
!     real*4        gammp 
      real          a,x,eps,gln,gold
      real          a0,a1,b0,b1,fac,an,ana,anf,g
      real          ap,sum,del,gamser,gammcf
      integer       itmax,n,ierr

      save

      data          rtname /'gammp'/
      data          itmax /1000/
      data          eps  /3.e-7/                

      ierr = 0
      gammp = 0.

      if (x.lt.a+1) then
        gln=gammln(a)
        ap=a
        sum=1./a
        del=sum
        do 111 n=1,itmax
          ap=ap+1.
          del=del*x/ap
          sum=sum+del
          if (abs(del).lt.abs(sum)*eps) go to 101
  111   continue
        ierr = 1 
!       call wrerr(rtname,ierr,'a too large, itmax too small ')

! 101   gamser = sum * exp( -x + a*log(x) - gln)
!       gammp = -alog(gamser)
  101   gammp = -alog(sum) + x - a*alog(x) + gln      
      else
        gln=gammln(a)
        gold=0.
        a0=1.
        a1=x
        b0=0.
        b1=1.
        fac=1.

        do 211 n=1,itmax
          an=float(n)
          ana=an-a
          a0=(a1+a0*ana)*fac
          b0=(b1+b0*ana)*fac
          anf=an*fac
          a1=x*a0+anf*a1
          b1=x*b0+anf*b1
          if (a1.ne.0) then
            fac=1./a1
            g=b1*fac
            if (abs((g-gold)/g).lt.eps) go to 201
            gold=g
          endif
  211   continue
        ierr = 1 
!       call wrerr(rtname,ierr,'a too large, itmax too small ')

  201   gammcf = exp(-x + a*alog(x) - gln) * g
        gammp = -alog(1.-gammcf)

      endif 

      end function gammp

!----------------------------------------------------------------
      real function gammln (xx)
!----------------------------------------------------------------

!  returns natural logarithm of the gamma function for xx>=1

!  author : s.doebereiner date: 11-sep-1996
!  returns natural logarithm of the gamma function for xx>=1
!  routine gammln from press et al., numerical recipes, 6.1, p.157

!  xx                  r4  i   argument
!  gammln              r4    o ln(gamma(xx))

!  wrflag              sr      writes to flag output stream
!  wrerr               sr      write error message

      character*8   rtname
      real          xx 
      double precision cof(6),stp,half,one,fpf,x,tmp,ser
      integer       j,ierr

      data          rtname /'gammln'/
      data cof,stp/76.18009173d0,-86.50532033d0,24.01409822d0, &
                   -1.231739516d0,.120858003d-2, &
                   -.536382d-5,2.50662827465d0/
      data half,one,fpf /0.5d0,1.0d0,5.5d0/
               
      ierr = 0
      gammln = 0.

      if (xx.le.0.0) then
        ierr = 1
!       call wrerr(rtname,ierr,'argument <=0 not allowed ') 
      endif

      if (ierr.eq.0) then 

! if xx<1 then use reflection formula (see press et al., eq. 6.1.4)     
        if (xx.lt.1) then
          x = xx + 1.d0 - one
        else
          x = xx - one
        endif       

        tmp=x+fpf
        tmp=(x+half)*log(tmp)-tmp
        ser=one
        do 111 j=1,6
          x=x+one
          ser=ser+cof(j)/x
  111   continue

! reflection formula for xx<1

!        if (xx.lt.1.) then
!          gammln=(tmp+log(stp*ser))/xx
!        else
!          gammln=tmp+log(stp*ser)        
!        endif 
!12/02/03: Changed to:

      gammln=tmp+log(stp*ser)
      if (xx.lt.1) gammln=gammln-log(xx)

      endif

      end function gammln

end module edetect_gen

