  module psf

  use param
  use dal
  use cal
  use emldata 
  use edetect_io_ml
  implicit none

  private
  public :: psf1, psf2, psf3, psf4, psf5, psf6

contains

  function psfoff (ene, eps, x, sig, ierr) result (result_psfoff) 

!  calculates XMM off-axis PSF (normalized surface brightness)
!
!  author : HEB        date: 15-NOV-1996 09:21 
!  modifi : GRH        date: 15-NOV-1996 20:07  off-axis from ABRIXAS 
!  modifi : GRH        date: 19-NOV-1996 09:50  XMM simulations included
!
!  The XMM QM PSF has been fit by Mike Watson with 4 Gaussians  
!  GRH has normalized the PSF components and added a mirror-blur term in
!  quadrature as a function of off axis angle, derived from XMM 
!  simulations performed by P.Friedrich using the ABRIXAS sim code.
!
!   call_var.          type I/O description
!  ENE                 R   I   Energy [keV]
!  EPS                  R   I      off-axis angle [arc min]
!  X                   R   I   angle from target position [arcsec]
!
!  IERR                I   O   = 0 no error 
!                              = 1 energy outside bounds 0.07-3 keV
!                              = 3 
!
!     OUTPUT:
!             PSFOFF      surface brightness of point spread function,
!                         normalized, so that  Integral 2*PI*r*dr*f
!                         from 0 to infinity = 1 [1/arcsec2]

    real          result_psfoff
    real          ene, eps, x, sig
    integer       ierr
    real          norm(4),sigma(4),sum,offsg,sig2,rnorm,sigtel
    integer       i

! MGW data norm  /0.75337,0.14463,0.79143e-2,0.75355e-3/,
! following are the norms after integrating each component to 1.00
    data norm  /0.4169,0.3470,0.1279,0.1082/, &
         sigma /4.1450,8.6218,22.372,66.698/, &
         offsg /0.0324/

    sum=0.

! off-axis component of telescope    
    sigtel=offsg*min(eps,18.0)**2
    do i=1,4
      sig2=0.5/(sigma(i)**2+sigtel**2+sig**2)
      rnorm=norm(i)*sig2/pi
      sum=sum+rnorm*exp(-sig2*(x)**2)
    end do

    result_psfoff = sum

  end function psfoff


!---------------------------------------------------------------------------------------

  real function psf1 (energy, iband, x1, y1, dx0, dy0, dsig, ecut,offax,azim,n) &
       result (result_psf1)

    real(kind=double), intent(in) ::  dx0, dy0, dsig
    real, intent(in) :: ecut
    real ::  x0, y0, sig
    real, dimension(max_ima), intent(in) :: energy
    real(kind=double), dimension(max_ins,max_fit)  :: offax,azim
    integer, intent(in) :: n
!    real  intent(in) :: ene
    integer, intent(in) :: iband, x1, y1
    integer :: j,k

    real, parameter :: sec2mm = 7500.0*pi/(180.0*3600.0)
    real(kind=double) :: E, theta, phi, val
    logical :: first = .true.
    integer :: i,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq
    integer, dimension(max_ima) :: nx, ny
    type (PsfT) :: ppsf
    type(PsfBinSizeT), dimension(max_ima) :: bsiz
    type(PSFwindowT) :: window
    type(PsfValidityRangesT) :: vrange 
    real(kind=single), dimension(:,:), pointer :: pima
    real, dimension(:,:,:), allocatable :: sima
!!    real, dimension(:,:,:), allocatable ::  psave
    real, dimension(0:100,0:100) :: gengen
    real :: x,y,dx,dy,ddx,ddy,xd0,yd0,xx,yy,x0old,y0old, &
            sum,winr,pfac,xyoff2,ecutold
    real :: sigold = -1.0
    integer :: sxlo,sxhi,sylo,syhi,nlo,nhi
    integer :: cl1,cl2,cl3,cl0

    save


    x0=sngl(dx0)
    y0=sngl(dy0)
    sig=sngl(dsig)


!    xd0=x0-xcen
!    yd0=y0-ycen
    xyoff2=(120.0/pixim)**2    

    if (first) then
!      count_model = 0
      allocate (sima(nim,0:250,0:250))
      nlo=30
      nhi=30
!!      allocate (psave(nim,nlo:nhi,nlo:nhi))
      gengen(0,0) = 1.0
    end if 

    if ((x0-x0old)**2+(y0-y0old)**2>xyoff2.or.ecut/=ecutold.or.first) then

      if (extpsf) then
        call CAL_setState1(instrument=EPN, &
                          AccuracyLevel=ACCURACY_EXTENDED)
      else
        call CAL_setState1(instrument=EPN, &
                          AccuracyLevel=ACCURACY_MEDIUM)
        psfnorm=1.0
      endif
!      write(*,*) "call CAL EPN"

     do j=1,num_instr
      theta=offax(j,n)     
      phi=azim(j,n)-pa_pnt(j)/180.*pi+pi/2.
      if (phi .gt. 2*pi) phi=phi-2.*pi
      if (phi .lt. -2.*pi) phi=phi+2.*pi 


      

         
      xd0=x0-xcen(j)
      yd0=y0-ycen(j)
!      theta=sqrt(xd0**2+yd0**2)*pixim
!      phi=atan2(xd0,yd0) + pi/2.0

      vrange = CAL_PsfValidityRanges()
      theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)

      np = max(0,nint((pixim-1.0)/2.0))
      np2= (2*np+1)**2 

      winr = int(ecut)+2.0
      winr=  min(winr*pixim,100.0)*sec2mm

      window%halfX = winr
      window%halfY = winr

!      write(*,*) "nim:",nim
!!!test write(*,*) i=1,nim 
!      do i=1,nim
      do k=1,num_image(j)
        i=ip(j,k) 
        E = energy(i)
        E = max(min(E,vrange%eMax-1.0),vrange%eMin+1.0)
        ppsf = Cal_getPSF(E, theta, phi) 
        bsiz(i) = CAL_psfBINSIZE(ppsf) 
        call CAL_psfImage (ppsf, window, pima) 
!       fudge factor for EXTENDED PSF:
        pima=pima*psfnorm

        nx(i) = size(pima,1)
        ny(i) = size(pima,2)
        do iy=np,ny(i)-1-np 
          do ix=np,nx(i)-1-np
            sum=0.0
            do ixx=ix-np,ix+np
              do iyy=iy-np,iy+np
                sum=sum+pima(ixx,iyy)
              end do
            end do
            sima(i,ix,iy)=sum/np2
          end do
        end do 
        call CAL_releaseMemory(pima)
      end do
     end do 

      ecutold=ecut
      first = .false.
!!      psave=-1.

    end if
 

!     count_model = count_model + 1

  
!!    ix=x1-int(x0)
!!    iy=y1-int(y0) 
!!    if (ix>=nlo.and.ix<=nhi.and.iy>=nlo.and.iy<=nhi) then
!!      write(*,*) "use psave"
!!      sum = psave(iband,ix,iy)
!!    else
!!      sum=-1.
!!    end if 
!!    if (x0==x0old.and.y0==y0old.and.sig==sigold.and.sum>0.) then
!!      result_psf1 = sum
!!      write(*,*) "use sum"
!!    else
      x0old=x0
      y0old=y0 

      xx = ((x1-x0)*pixim*sec2mm+window%halfX)/bsiz(iband)%binX-0.5
      yy = ((y1-y0)*pixim*sec2mm+window%halfY)/bsiz(iband)%binY-0.5   

      pfac=pixim*sec2mm/bsiz(iband)%binX
! extend calculation to 10 sigma (06/11/05)
      isig3 = nint(min(10.*sig,ecut))
      isig3sq = isig3**2



      if (sig == 0.0) then  
        gengen=0.0
        gengen(0,0)=1.0     
      endif
      if (sig/=sigold .and. sig > 0.0) then
        do ly = 0,isig3
          y=ly
          do lx = 0,ly
            if (lx**2+ly**2<=isig3sq) then        
              x=lx
              if (extmod == "GAUSSIAN") then
!               Gaussian extent:
                if (lx/=ly) then
                  gengen(lx,ly) = gen(0.0,x-0.5,x+0.5,sig)* &
                                  gen(0.0,y-0.5,y+0.5,sig)
                else
                  gengen(lx,lx) = gen(0.0,x-0.5,x+0.5,sig)**2
                end if
              else
!               Beta model:
                gengen(lx,ly)=king(x,y,sig) 
              end if
            end if 
          end do
        end do
      end if
      sigold=sig


      sxlo = max(-isig3,int((1-xx)/pfac))
      sxhi = min(isig3,int((nx(iband)-2-xx)/pfac)+1)
      sylo = max(-isig3,int((1-yy)/pfac))
      syhi = min(isig3,int((ny(iband)-2-yy)/pfac)+1)

      sum = 0.
      do ly = sylo,syhi   ! -isig3,isig3 
        do lx =  sxlo,sxhi ! -isig3,isig3 
          x=xx+lx*pfac
          y=yy+ly*pfac 
          ix = min(max(int(x),np),nx(iband)-2-np)
          iy = min(max(int(y),np),ny(iband)-2-np)
          dx=max(x-ix,0.)
          dy=max(y-iy,0.)
          ddx=max(1.-dx,0.)
          ddy=max(1.-dy,0.)

          val = ddx*ddy*sima(iband,ix  ,iy)   + &
                ddx* dy*sima(iband,ix  ,iy+1) + &
                 dx*ddy*sima(iband,ix+1,iy)   + &
                 dx* dy*sima(iband,ix+1,iy+1)
        
          kx = abs(lx)
          ky = abs(ly)
          if (kx>ky) then
            kk=kx
            kx=ky
            ky=kk
          end if 
          sum = sum + val * gengen (kx,ky)

        end do
      end do

      sum = sum * (sec2mm*pixim)**2
      result_psf1 = sum

!!      ix=x1-int(x0)
!!      iy=y1-int(y0) 
!!      if (ix>=nlo.and.ix<=nhi.and.iy>=nlo.and.iy<=nhi) &
!!        psave(iband,ix,iy) = sum    

!!    end if

  call system_clock(cl1,cl2,cl3)
!  write(*,*) "psf time:", float(cl1-cl0)/1000.

  end function psf1

!---------------------------------------------------------------------------------------

  real function psf2 (energy, iband, x1, y1, dx0, dy0, dsig, ecut,offax,azim,n) &
       result (result_psf2)

    real(kind=double), intent(in) ::  dx0, dy0, dsig
    real, intent(in) :: ecut
    real ::  x0, y0, sig
    real, dimension(max_ima), intent(in) :: energy
    real(kind=double), dimension(max_ins,max_fit)  :: offax,azim
    integer, intent(in) :: n
    integer :: iband, x1, y1
    integer :: j,k
    real, parameter :: sec2mm = 7500.0*pi/(180.0*3600.0)
    real(kind=double) :: E, theta, phi, val
    logical :: first = .true.
    integer :: i,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq
    integer, dimension(max_ima) :: nx, ny
    type (PsfT) :: ppsf
    type(PsfBinSizeT), dimension(max_ima) :: bsiz
    type(PSFwindowT) :: window
    type(PsfValidityRangesT) :: vrange 
    real(kind=single), dimension(:,:), pointer :: pima
    real, dimension(:,:,:), allocatable :: sima
!!    real, dimension(:,:,:), allocatable :: psave 
    real, dimension(0:100,0:100) :: gengen
    real :: x,y,dx,dy,ddx,ddy,xd0,yd0,xx,yy,x0old,y0old, &
            sum,winr,pfac,xyoff2,ecutold
    real :: sigold = -1.0
    integer :: sxlo,sxhi,sylo,syhi,nlo,nhi

    save


    x0=sngl(dx0)
    y0=sngl(dy0)
    sig=sngl(dsig)


!    xd0=x0-xcen
!    yd0=y0-ycen
    xyoff2=(120.0/pixim)**2    

    if (first) then
      allocate (sima(nim,0:250,0:250))
      nlo=30
      nhi=30
!!      allocate (psave(nim,nlo:nhi,nlo:nhi))
      gengen(0,0) = 1.0
    end if 

    if ((x0-x0old)**2+(y0-y0old)**2>xyoff2.or.ecut/=ecutold.or.first) then


    if (extpsf) then
      call CAL_setState1(instrument=EMOS1, &
                         AccuracyLevel=ACCURACY_EXTENDED)
    else
      call CAL_setState1(instrument=EMOS1, &
                         AccuracyLevel=ACCURACY_MEDIUM)
      psfnorm=1.0
    endif

!      write(*,*) "call CALL EMOS1"
     do j=1,num_instr
      theta=offax(j,n)     
      phi=azim(j,n)-pa_pnt(j)/180.*pi+pi/2.
      if (phi .gt. 2.*pi) phi=phi-2.*pi
      if (phi .lt. -2.*pi) phi=phi+2.*pi
!      xd0=x0-xcen(j)
!      yd0=y0-ycen(j)
!      theta=sqrt(xd0**2+yd0**2)*pixim
!      phi=atan2(xd0,yd0) + pi/2.0

      vrange = CAL_PsfValidityRanges()
      theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)

      np = max(0,nint((pixim-1.0)/2.0))
      np2= (2*np+1)**2 

      winr = int(ecut)+2.0
      winr=  min(winr*pixim,100.0)*sec2mm

      window%halfX = winr
      window%halfY = winr

!      do i=1,nim
      do k=1,num_image(j)
        i=ip(j,k)
        E = energy(i)
        E = max(min(E,vrange%eMax-1.0),vrange%eMin+1.0)
        ppsf = Cal_getPSF(E, theta, phi) 
        bsiz(i) = CAL_psfBINSIZE(ppsf) 
        call CAL_psfImage (ppsf, window, pima)
!       fudge factor for EXTENDED PSF:
        pima=pima*psfnorm
 
        nx(i) = size(pima,1)
        ny(i) = size(pima,2)
        do iy=np,ny(i)-1-np 
          do ix=np,nx(i)-1-np
            sum=0.0
            do ixx=ix-np,ix+np
              do iyy=iy-np,iy+np
                sum=sum+pima(ixx,iyy)
              end do
            end do
            sima(i,ix,iy)=sum/np2
          end do
        end do 
        call CAL_releaseMemory(pima)
      end do 
     end do
     ecutold=ecut
     first = .false. 
!!      psave=-1.

    end if

!    count_model = count_model + 1

!!    ix=x1-int(x0)
!!    iy=y1-int(y0) 
!!    if (ix>=nlo.and.ix<=nhi.and.iy>=nlo.and.iy<=nhi) then
!!      sum = psave(iband,ix,iy)
!!    else
!!      sum=-1.
!!    end if 
!!    if (x0==x0old.and.y0==y0old.and.sig==sigold.and.sum>0.) then
!!     write(*,*) "use saved PSF"
!!      result_psf2 = sum     
!!    else
      x0old=x0
      y0old=y0 

      xx = ((x1-x0)*pixim*sec2mm+window%halfX)/bsiz(iband)%binX-0.5
      yy = ((y1-y0)*pixim*sec2mm+window%halfY)/bsiz(iband)%binY-0.5   

      pfac=pixim*sec2mm/bsiz(iband)%binX
! extend calculation to 10 sigma (06/11/05)
      isig3 = nint(min(10.*sig,ecut))
      isig3sq = isig3**2

      if (sig == 0.0) then  
        gengen=0.0
        gengen(0,0)=1.0     
      endif
      if (sig/=sigold .and. sig > 0.0) then
        do ly = 0,isig3
          do lx = 0,ly
            if (lx**2+ly**2<=isig3sq) then        
              x=lx
              y=ly
              if (extmod == "GAUSSIAN") then 
!               Gaussian extent:
              if (lx/=ly) then
                gengen(lx,ly) = gen(0.0,x-0.5,x+0.5,sig)* &
                                gen(0.0,y-0.5,y+0.5,sig)
              else
                gengen(lx,lx) = gen(0.0,x-0.5,x+0.5,sig)**2
              end if
              else
!               Beta model:
                gengen(lx,ly)=king(x,y,sig) 
              end if
            end if 
          end do
        end do
      end if
      sigold=sig

      sxlo = max(-isig3,int((1-xx)/pfac))
      sxhi = min(isig3,int((nx(iband)-2-xx)/pfac)+1)
      sylo = max(-isig3,int((1-yy)/pfac))
      syhi = min(isig3,int((ny(iband)-2-yy)/pfac)+1)

      sum = 0.
      do ly = sylo,syhi    ! -isig3,isig3
        do lx = sxlo,sxhi  ! -isig3,isig3
          x=xx+lx*pfac
          y=yy+ly*pfac 
          ix = min(max(int(x),np),nx(iband)-2-np)
          iy = min(max(int(y),np),ny(iband)-2-np)
          dx=max(x-ix,0.)
          dy=max(y-iy,0.)
          ddx=max(1.-dx,0.)
          ddy=max(1.-dy,0.)

          val = ddx*ddy*sima(iband,ix  ,iy)   + &
                ddx* dy*sima(iband,ix  ,iy+1) + &
                 dx*ddy*sima(iband,ix+1,iy)   + &
                 dx* dy*sima(iband,ix+1,iy+1)
        
          kx = abs(lx)
          ky = abs(ly)
          if (kx>ky) then
            kk=kx
            kx=ky
            ky=kk
          end if 
          sum = sum + val * gengen (kx,ky)

        end do
      end do 
    
      sum = sum * (sec2mm*pixim)**2
      result_psf2 = sum

!!      ix=x1-int(x0)
!!      iy=y1-int(y0) 
!!      if (ix>=nlo.and.ix<=nhi.and.iy>=nlo.and.iy<=nhi) &
!!        psave(iband,ix,iy) = sum    

!!    end if

  end function psf2

!---------------------------------------------------------------------------------------

  real function psf3 (energy, iband, x1, y1, dx0, dy0, dsig, ecut,offax,azim,n) &
       result (result_psf3)

    real(kind=double), intent(in) ::  dx0, dy0, dsig
    real, intent(in) :: ecut
    real ::  x0, y0, sig
    real, dimension(max_ima), intent(in) :: energy
    real(kind=double), dimension(max_ins,max_fit)  :: offax,azim
    integer, intent(in) :: n
    integer :: iband, x1, y1
    integer :: j,k
    real, parameter :: sec2mm = 7500.0*pi/(180.0*3600.0)
    real(kind=double) :: E, theta, phi, val
    logical :: first = .true.
    integer :: i,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq
    integer, dimension(max_ima) :: nx, ny
    type (PsfT) :: ppsf
    type(PsfBinSizeT), dimension(max_ima) :: bsiz
    type(PSFwindowT) :: window
    type(PsfValidityRangesT) :: vrange 
    real(kind=single), dimension(:,:), pointer :: pima
    real, dimension(:,:,:), allocatable :: sima
!!    real, dimension(:,:,:), allocatable :: psave 
    real, dimension(0:100,0:100) :: gengen
    real :: x,y,dx,dy,ddx,ddy,xd0,yd0,xx,yy,x0old,y0old, &
            sum,winr,pfac,xyoff2,ecutold,testsum
    real :: sigold = -1.0
    integer :: sxlo,sxhi,sylo,syhi,nlo,nhi

    save

    x0=sngl(dx0)
    y0=sngl(dy0)
    sig=sngl(dsig)


!    xd0=x0-xcen
!    yd0=y0-ycen
    xyoff2=(120.0/pixim)**2    

    if (first) then
      allocate (sima(nim,0:250,0:250))
      nlo=30
      nhi=30
!!      allocate (psave(nim,nlo:nhi,nlo:nhi))
      gengen(0,0) = 1.0
    end if 

    if ((x0-x0old)**2+(y0-y0old)**2>xyoff2.or.ecut/=ecutold.or.first) then


      if (extpsf) then
        call CAL_setState1(instrument=EMOS2, &
                           AccuracyLevel=ACCURACY_EXTENDED)
      else
        call CAL_setState1(instrument=EMOS2, &
                           AccuracyLevel=ACCURACY_MEDIUM)
        psfnorm=1.0
      endif

     do j=1,num_instr
       theta=offax(j,n)     
       phi=azim(j,n)-pa_pnt(j)/180.*pi+pi/2. 
       if (phi .gt. 2*pi) phi=phi-2.*pi
       if (phi .lt. -2.*pi) phi=phi+2.*pi 
!      xd0=x0-xcen(j)
!      yd0=y0-ycen(j)
!      theta=sqrt(xd0**2+yd0**2)*pixim
!      phi=atan2(xd0,yd0) + pi/2.0

      vrange = CAL_PsfValidityRanges()
      theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)

      np = max(0,nint((pixim-1.0)/2.0))
      np2= (2*np+1)**2 

      winr = int(ecut)+2.0
      winr=  min(winr*pixim,100.0)*sec2mm

      window%halfX = winr
      window%halfY = winr

!      do i=1,nim
      do k=1,num_image(j)

        i=ip(j,k)
        E = energy(i)
        E = max(min(E,vrange%eMax-1.0),vrange%eMin+1.0)
        ppsf = Cal_getPSF(E, theta, phi) 
        bsiz(i) = CAL_psfBINSIZE(ppsf) 
        call CAL_psfImage (ppsf, window, pima) 
!       fudge factor for EXTENDED PSF:
        pima=pima*psfnorm

        nx(i) = size(pima,1)
        ny(i) = size(pima,2)
        do iy=np,ny(i)-1-np 
          do ix=np,nx(i)-1-np
            sum=0.0
            do ixx=ix-np,ix+np
              do iyy=iy-np,iy+np
                sum=sum+pima(ixx,iyy)
              end do
            end do
            sima(i,ix,iy)=sum/np2
          end do
        end do 
        call CAL_releaseMemory(pima)
      end do
     end do 

     ecutold=ecut
     first = .false. 
!!   psave=-1.

    end if

!!    ix=x1-int(x0)
!!    iy=y1-int(y0) 
!!    if (ix>=nlo.and.ix<=nhi.and.iy>=nlo.and.iy<=nhi) then
!!      sum = psave(iband,ix,iy)
!!    else
!!      sum=-1.
!!    end if 
!!    if (x0==x0old.and.y0==y0old.and.sig==sigold.and.sum>0.) then
!!      result_psf3 = sum 
!!    else
      x0old=x0
      y0old=y0 

      xx = ((x1-x0)*pixim*sec2mm+window%halfX)/bsiz(iband)%binX-0.5
      yy = ((y1-y0)*pixim*sec2mm+window%halfY)/bsiz(iband)%binY-0.5   
      pfac=pixim*sec2mm/bsiz(iband)%binX
! extend calculation to 10 sigma (06/11/05)
      isig3 = nint(min(10.*sig,ecut))
      isig3sq = isig3**2

      if (sig == 0.0) then  
        gengen=0.0
        gengen(0,0)=1.0     
      endif
      if (sig/=sigold .and. sig > 0.0) then
        testsum=0.0
        do ly = 0,isig3
          do lx = 0,ly
            if (lx**2+ly**2<=isig3sq) then        
              x=lx
              y=ly
              if (extmod == "GAUSSIAN") then 
!             Gaussian extent:
              if (lx/=ly) then
                gengen(lx,ly) = gen(0.0,x-0.5,x+0.5,sig)* &
                                gen(0.0,y-0.5,y+0.5,sig)
              else
                gengen(lx,lx) = gen(0.0,x-0.5,x+0.5,sig)**2
              end if
              else
!               Beta model:
                gengen(lx,ly)=king(x,y,sig) 
              end if
            end if 
          end do
        end do
      end if
      sigold=sig
      sum = 0.

      sxlo = max(-isig3,int((1-xx)/pfac))
      sxhi = min(isig3,int((nx(iband)-2-xx)/pfac)+1)
      sylo = max(-isig3,int((1-yy)/pfac))
      syhi = min(isig3,int((ny(iband)-2-yy)/pfac)+1)

      do ly = sylo,syhi   ! -isig3,isig3
        do lx = sxlo,sxhi ! -isig3,isig3
          x=xx+lx*pfac
          y=yy+ly*pfac 
          ix = min(max(int(x),np),nx(iband)-2-np)
          iy = min(max(int(y),np),ny(iband)-2-np)
          dx=max(x-ix,0.)
          dy=max(y-iy,0.)
          ddx=max(1.-dx,0.)
          ddy=max(1.-dy,0.)

          val = ddx*ddy*sima(iband,ix  ,iy)   + &
                ddx* dy*sima(iband,ix  ,iy+1) + &
                 dx*ddy*sima(iband,ix+1,iy)   + &
                 dx* dy*sima(iband,ix+1,iy+1)
        
          kx = abs(lx)
          ky = abs(ly)
          if (kx>ky) then
            kk=kx
            kx=ky
            ky=kk
          end if 
          sum = sum + val * gengen (kx,ky)

        ! end if
        end do
      end do 
    
      sum = sum * (sec2mm*pixim)**2
      result_psf3 = sum

!!      ix=x1-int(x0)
!!      iy=y1-int(y0) 
!!      if (ix>=nlo.and.ix<=nhi.and.iy>=nlo.and.iy<=nhi) &
 !!       psave(iband,ix,iy) = sum    

 !!   end if

  end function psf3

!---------------------------------------------------------------------------------------

  real function psf4 (energy, iband, x1, y1, dx0, dy0, xcen, ycen, dsig, ecut,theta) &
       result (result_psf4)


    real(kind=double), intent(in) ::  dx0, dy0, dsig
    real, intent(in) :: ecut,xcen,ycen
    real ::  x0, y0, sig
    real(kind=double), intent(in) :: theta
    real, dimension(max_ima), intent(in) :: energy
    integer, intent(in) :: iband
    real ::  x1, y1
    integer :: ierr

    real    :: x, eps, xsig, xd0, yd0, rc2, r2
       
  
    x0=sngl(dx0)
    y0=sngl(dy0)
    sig=sngl(dsig)



    xd0=x0-xcen
    yd0=y0-ycen
      
    if (extmod == "GAUSSIAN" .or. sig .lt. 2.0) then 
      x = sqrt ((x1-x0)**2+(y1-y0)**2) * pixim
      xsig = sig * pixim
!    eps = sqrt(xd0**2+yd0**2)*pixim/60.
      eps = theta/60.
      result_psf4 = psfoff (1.0, eps, x, xsig, ierr) * pixim**2
    else
      !beta model:
      rc2=sig*sig
      r2=(x1-x0)*(x1-x0)+(y1-y0)*(y1-y0)
      result_psf4=(1+r2/rc2)**(-1.5)/(twopi*rc2)
    endif
   
  end function psf4
!---------------------------------------------------------------------------------------

  real function psf5 (ene, offax, posang, x1, y1, x0, y0, sig)

    real, intent(in) :: ene, offax, posang, x0, y0, x1, y1, sig
    psf5 = 0.

  end function psf5

!---------------------------------------------------------------------------------------
  real function psf6 (ene,offax,posang,x1,y1,x0,y0,sig)

    real, intent(in) :: ene, offax, posang, x0, y0, x1, y1, sig
    psf6 = 0.

  end function psf6

!---------------------------------------------------------------------------------------

  real function gen(x0,x1,x2,sig)

    real x0,x1,x2,sig,fsig
                                                                       
    gen=0.
    fsig=sig                                                           
    if (fsig.le.0.00001) fsig=0.00001
    fsig = 1.0/(1.414214*fsig)
    gen = (errf((x2-x0)*fsig)-errf((x1-x0)*fsig))/2.0

  end function gen     

!---------------------------------------------------------------------------------------

  real function errf (xx)

    real t,x,xx

    x=abs(xx)                                                        
    errf=1.0
                                                         
    if (x<=5.0) then                                            
       t=1.0/(1.0+0.47047*x)                                            
       errf=1.0-(0.3480242-(0.0958798-0.7478556*t)*t)*t*exp(-(x*x))
     end if   
                                                                    
     if (xx<0.0) errf=-errf                                        

   end function errf     

!---------------------------------------------------------------------------------------
   real function king(x,y,rcore)

!  returns value of the King profile (1+r**2/rcore**2)**(-1.5) integrated over the pixel with center (x,y)
!  normalized to 1  
!  the integration is approximated by the radial integral between r1 and r2, r1=r_pix-0.5, r2=r_pix+0.5

     real x,y,rcore
     real r1,r2,r1sq,r2sq,rc2,r
     real kingint,area,norm


     rc2=max(1E-2,rcore*rcore)
  
     r=(x*x+y*y)**0.5
     r1=max(0.,r-0.5)
     r2=r+0.5    
     r1sq=r1*r1
     r2sq=r2*r2 
!-----------------------------------------------------
!     kingint=twopi*(rc2/(1+r1sq/rc2)-rc2/(1+r2sq/rc2))
!     area=pi*(r2sq-r1sq)
!     norm=twopi*rc2
!     king=kingint/area/norm 
!----------------------------------------------------

!    Correct integration of king model (26/06/05 GL):
     norm=twopi*rc2
     king=(r2/sqrt(1.+r2sq/rc2)-(r1/sqrt(1.+r1sq/rc2)))/(r2-r1)/norm
 

   end function king

!---------------------------------------------------------------------------------------

end module psf





